use cgmath::{Deg, EuclideanSpace, InnerSpace, Matrix2, Point2, Rad, Vector2, Zero};
use crossbeam::channel::Receiver;
use nalgebra::{point, vector};
use object::{ship::Ship, GameData, GameShipHandle};
use rand::Rng;
use rapier2d::{
	dynamics::{RigidBody, RigidBodyBuilder, RigidBodyHandle},
	geometry::{ColliderBuilder, ColliderHandle, CollisionEvent},
	pipeline::{ActiveEvents, ChannelEventCollector},
};
use std::{collections::HashMap, f32::consts::PI};

use crate::{
	behavior, objects,
	objects::{ProjectileWorldObject, ShipWorldObject},
	util,
	wrapper::Wrapper,
	ParticleBuilder, StepOutput, StepResources,
};
use galactica_content as content;
use galactica_gameobject as object;

/// Manages the physics state of one system
pub struct World {
	/// The system this world is attached to
	system: content::SystemHandle,

	wrapper: Wrapper,
	projectiles: HashMap<ColliderHandle, objects::ProjectileWorldObject>,
	ships: HashMap<GameShipHandle, objects::ShipWorldObject>,
	collider_ship_table: HashMap<ColliderHandle, GameShipHandle>,

	collision_handler: ChannelEventCollector,
	collision_queue: Receiver<CollisionEvent>,
}

// Private methods
impl<'a> World {
	fn remove_projectile(
		&mut self,
		c: ColliderHandle,
	) -> Option<(RigidBody, ProjectileWorldObject)> {
		let p = match self.projectiles.remove(&c) {
			Some(p) => p,
			None => return None,
		};
		let r = self
			.wrapper
			.rigid_body_set
			.remove(
				p.rigid_body,
				&mut self.wrapper.im,
				&mut self.wrapper.collider_set,
				&mut self.wrapper.ij,
				&mut self.wrapper.mj,
				true,
			)
			.unwrap();

		return Some((r, p));
	}

	fn remove_ship(&mut self, s: &ShipWorldObject) {
		self.wrapper.rigid_body_set.remove(
			s.rigid_body,
			&mut self.wrapper.im,
			&mut self.wrapper.collider_set,
			&mut self.wrapper.ij,
			&mut self.wrapper.mj,
			true,
		);
		let h = self.collider_ship_table.remove(&s.collider).unwrap();
		self.ships.remove(&h);
	}

	/*
	/// Add a projectile fired from a ship
	fn add_projectiles(
		&mut self,
		s: ShipPhysicsHandle,
		p: Vec<(ProjectileWorldObject, content::GunPoint)>,
	) {
		let mut rng = rand::thread_rng();
		for (projectile, point) in p {
			let r = self.get_ship_body(s).unwrap().1;

			let ship_pos = util::rigidbody_position(r);
			let ship_rot = util::rigidbody_rotation(r);
			let ship_vel = util::rigidbody_velocity(r);
			let ship_ang = ship_rot.angle(Vector2 { x: 0.0, y: 1.0 });

			let pos = ship_pos + (Matrix2::from_angle(-ship_ang) * point.pos.to_vec());

			let spread: Rad<f32> = Deg(
				rng.gen_range(-projectile.content.angle_rng.0..=projectile.content.angle_rng.0)
			)
			.into();

			let vel = ship_vel
				+ (Matrix2::from_angle(-ship_ang + spread)
					* Vector2 {
						x: 0.0,
						y: projectile.content.speed
							+ rng.gen_range(
								-projectile.content.speed_rng..=projectile.content.speed_rng,
							),
					});

			let rigid_body = RigidBodyBuilder::kinematic_velocity_based()
				.translation(vector![pos.x, pos.y])
				.rotation(-ship_ang.0)
				.linvel(vector![vel.x, vel.y])
				.build();

			let collider = match &projectile.content.collider {
				content::ProjectileCollider::Ball(b) => ColliderBuilder::ball(b.radius)
					.sensor(true)
					.active_events(ActiveEvents::COLLISION_EVENTS)
					.build(),
			};

			let rigid_body = self.wrapper.rigid_body_set.insert(rigid_body);
			let collider = self.wrapper.collider_set.insert_with_parent(
				collider,
				rigid_body,
				&mut self.wrapper.rigid_body_set,
			);

			self.projectiles.insert(
				collider.clone(),
				ProjectileWorldObject::new(projectile, rigid_body, collider),
			);
		}
	}
	*/

	fn collide_projectile_ship(
		&mut self,
		res: &mut StepResources,
		projectile_h: ColliderHandle,
		ship_h: ColliderHandle,
	) {
		let projectile = self.projectiles.get(&projectile_h);
		let ship = self
			.ships
			.get_mut(self.collider_ship_table.get(&ship_h).unwrap());
		if projectile.is_none() || ship.is_none() {
			return;
		}
		let projectile = projectile.unwrap();
		let ship = ship.unwrap();

		let ship_d = res.dt.get_ship_mut(ship.data_handle).unwrap();

		// TODO: check faction
		ship_d.apply_damage(projectile.content.damage);

		if true {
			let pr = self
				.wrapper
				.rigid_body_set
				.get(projectile.rigid_body)
				.unwrap();
			let v = util::rigidbody_velocity(pr).normalize() * projectile.content.force;
			let pos = util::rigidbody_position(pr);
			let _ = pr;

			let r = self
				.wrapper
				.rigid_body_set
				.get_mut(ship.rigid_body)
				.unwrap();
			r.apply_impulse_at_point(vector![v.x, v.y], point![pos.x, pos.y], true);

			// Borrow again, we can only have one at a time
			let pr = self
				.wrapper
				.rigid_body_set
				.get(projectile.rigid_body)
				.unwrap();
			let pos = util::rigidbody_position(pr);
			let angle = util::rigidbody_rotation(pr).angle(Vector2 { x: 1.0, y: 0.0 });

			match &projectile.content.impact_effect {
				None => {}
				Some(x) => {
					let effect = res.ct.get_effect(*x);
					let r = ship.rigid_body;
					let sr = self.get_rigid_body(r).unwrap();
					let parent_velocity = util::rigidbody_velocity(pr);
					let target_velocity =
						sr.velocity_at_point(&nalgebra::Point2::new(pos.x, pos.y));

					res.particles.push(ParticleBuilder::from_content(
						effect,
						pos,
						-angle,
						parent_velocity,
						Vector2 {
							x: target_velocity.x,
							y: target_velocity.y,
						},
					));
				}
			};

			self.remove_projectile(projectile.collider);
		}
	}
}

// Public methods
impl<'a> World {
	/// Create a new physics system
	pub fn new(ct: &content::Content, dt: &GameData, system: content::SystemHandle) -> Self {
		let (collision_send, collision_queue) = crossbeam::channel::unbounded();
		let (contact_force_send, _) = crossbeam::channel::unbounded();

		let mut w = Self {
			system,
			wrapper: Wrapper::new(),
			projectiles: HashMap::new(),
			ships: HashMap::new(),
			collider_ship_table: HashMap::new(),
			collision_handler: ChannelEventCollector::new(collision_send, contact_force_send),
			collision_queue,
		};

		// TODO: guarantee not touching
		// TODO: add, remove ships each tick
		// Maybe store position in gamedata?
		let mut rng = rand::thread_rng();
		for s in dt.iter_ships() {
			w.add_ship(
				ct,
				s,
				Point2 {
					x: rng.gen_range(-500.0..=500.0),
					y: rng.gen_range(-500.0..=500.0),
				},
			);
		}

		return w;
	}

	/// Add a ship to this physics system
	pub fn add_ship(&mut self, ct: &content::Content, ship: &Ship, position: Point2<f32>) {
		let ship_content = ct.get_ship(ship.get_content());
		let cl = ColliderBuilder::convex_decomposition(
			&ship_content.collision.points[..],
			&ship_content.collision.indices[..],
		)
		// Rotate collider to match sprite
		// (Collider starts pointing east, sprite starts pointing north.)
		.rotation(PI / -2.0)
		.mass(ship_content.mass);
		// TODO: only build colliders once

		let rb = RigidBodyBuilder::dynamic()
			.angular_damping(ship_content.angular_drag)
			.linear_damping(ship_content.linear_drag)
			.translation(vector![position.x, position.y])
			.can_sleep(false);

		let r = self.wrapper.rigid_body_set.insert(rb.build());
		let c = self.wrapper.collider_set.insert_with_parent(
			cl.build(),
			r,
			&mut self.wrapper.rigid_body_set,
		);

		self.collider_ship_table.insert(c, ship.get_handle());
		self.ships.insert(
			ship.get_handle(),
			objects::ShipWorldObject::new(
				ct,
				ship.get_handle(),
				Box::new(behavior::Null::new()),
				r,
				c,
			),
		);
	}

	/// Step this physics system by `t` seconds
	pub fn step(&mut self, mut res: StepResources) -> StepOutput {
		let mut output = StepOutput {
			player_position: Point2 { x: 0.0, y: 0.0 },
		};

		// Run ship updates
		// TODO: maybe reorganize projectile creation?
		//let mut projectiles = Vec::new();
		let mut to_remove = Vec::new();
		for (_, s) in &mut self.ships {
			let r = &mut self.wrapper.rigid_body_set[s.rigid_body];
			let c = &mut self.wrapper.collider_set[s.collider];

			if s.data_handle == res.player {
				s.controls = res.player_controls.clone();
				output.player_position = util::rigidbody_position(r);
			} else {
				s.update_controls(&res);
			}

			// TODO: unified step info struct
			s.step(&mut res, r, c);
			//if s.controls.guns {
			//	projectiles.push((s.physics_handle, s.ship.fire_guns()));
			//}

			//if s.remove_from_world() {
			//	to_remove.push(s.physics_handle);
			//	continue;
			//}
		}
		//for (s, p) in projectiles {
		//	self.add_projectiles(s, p);
		//}
		for s in to_remove {
			self.remove_ship(s);
		}

		// Update physics
		self.wrapper.step(res.t, &self.collision_handler);

		// Handle collision events
		while let Ok(event) = &self.collision_queue.try_recv() {
			if event.started() {
				let a = event.collider1();
				let b = event.collider2();

				// If projectiles are part of this collision, make sure
				// `a` is one of them.
				let (a, b) = if self.projectiles.contains_key(&b) {
					(b, a)
				} else {
					(a, b)
				};

				let p = self.projectiles.get(&a);
				let s = self
					.ships
					.get_mut(self.collider_ship_table.get(&b).unwrap());
				if p.is_none() || s.is_none() {
					continue;
				}
				self.collide_projectile_ship(&mut res, a, b);
			}
		}

		// Delete projectiles
		let mut to_remove = Vec::new();
		for (c, p) in &mut self.projectiles {
			p.tick(res.t);
			if p.is_expired() {
				to_remove.push(*c);
			}
		}

		let mut rng = rand::thread_rng();
		for c in to_remove {
			let (pr, p) = self.remove_projectile(c).unwrap();

			match &p.content.expire_effect {
				None => {}
				Some(x) => {
					let x = res.ct.get_effect(*x);
					let pos = util::rigidbody_position(&pr);
					let vel = util::rigidbody_velocity(&pr);
					let angle = util::rigidbody_rotation(&pr).angle(Vector2 { x: 1.0, y: 0.0 });

					let velocity = {
						let a = rng
							.gen_range(-x.velocity_scale_parent_rng..=x.velocity_scale_parent_rng);

						let velocity = (x.velocity_scale_parent + a) * vel;

						velocity
					};

					res.particles.push(ParticleBuilder::from_content(
						x,
						pos,
						-angle,
						velocity,
						Vector2::zero(),
					));
				}
			};
		}

		return output;
	}

	pub fn get_ship(&self, ship: GameShipHandle) -> Option<&ShipWorldObject> {
		self.ships.get(&ship)
	}

	/// Get a rigid body from a handle
	pub fn get_rigid_body(&self, r: RigidBodyHandle) -> Option<&RigidBody> {
		self.wrapper.rigid_body_set.get(r)
	}

	/// Get a rigid body from a handle
	pub fn get_rigid_body_mut(&mut self, r: RigidBodyHandle) -> Option<&mut RigidBody> {
		self.wrapper.rigid_body_set.get_mut(r)
	}

	/// Iterate over all ships in this physics system
	pub fn iter_ship_body(
		&self,
	) -> impl Iterator<Item = (&objects::ShipWorldObject, &RigidBody)> + '_ {
		self.ships
			.values()
			.map(|x| (x, self.wrapper.rigid_body_set.get(x.rigid_body).unwrap()))
	}

	/// Iterate over all ships in this physics system
	pub fn iter_ships(&'a self) -> impl Iterator<Item = &ShipWorldObject> + '_ {
		self.ships.values()
	}

	/// Iterate over all ships in this physics system
	pub fn iter_projectiles(&'a self) -> impl Iterator<Item = &ProjectileWorldObject> + '_ {
		self.projectiles.values()
	}
}
